import numpy as np
from sklearn.model_selection import train_test_split
from sklearn.ensemble import RandomForestClassifier
from annoy import AnnoyIndex

import scipy.spatial
import scipy.stats
import sklearn.metrics

from annoy import AnnoyIndex
import itertools
from typing import Optional, Any, Union, List, Tuple, Callable
import warnings


class EvaluationMetrics():
    """Evaluation metrics for dimension reduction
    
    This class contains methods to run evluation metrics.
    """
        
    @staticmethod
    def build_annoy(data: "np.ndarray",
                    saved_annoy_path: Optional[str]=None,
                    k: int=5) -> "np.ndarray":
        '''Build ANNOY and returns nearest neighbors.
        
        This is a utility function for building ANNOY models and returning the nearest-neighbor matrices for original
        space data and low-dimensional embedding. 
        
        
        : param data: The input high-dimensional array.
        : param saved_annoy_path: The path to pre-built ANNOY model for original data.
        : param k: The number of neighbors.

        :return: Nearest-neighbor matrices of original space data.
        '''
        
        if saved_annoy_path is not None:
            annoy_model_data = Annoy.load_annoy(path=saved_annoy_path, ncol=data.shape[1])
        else:
            annoy_model_data = Annoy.build_annoy(data)
            
        annoy_data_neighbors = np.empty((data.shape[0], k), dtype=int)
        for i in range(data.shape[0]):
            data_neighbors: List[int] = annoy_model_data.get_nns_by_item(i, k+1)
            # Try remove the i-th obs itself
            try:
                data_neighbors.remove(i)
            except ValueError: # pragma: no cover
                data_neighbors = data_neighbors[0:(len(data_neighbors)-1)]
                
            annoy_data_neighbors[i] = data_neighbors
        
        return annoy_data_neighbors
        
    
    @staticmethod
    def correlation(x: "np.ndarray",
                    y: "np.ndarray",
                    metric: str="Pearson") -> float:
        '''Calculate Correlation Coefficient
        
        This method computes the pearson or spearman correlation between the inputs.

        :param x: The first 1D array. 
        :param y: The second 1D array.
        :param metric: The metric to use. 'Pearson' or 'Spearman', defaults to "Pearson".

        :return: Correlation coefficient.
        '''
        if metric.lower() == "pearson":
            cor: float=scipy.stats.pearsonr(x, y)[0]
        elif metric.lower() == "spearman":
            cor: float=scipy.stats.spearmanr(x, y)[0]
        else:
            raise ValueError("Unsupported metric: must be 'Pearson' or 'Spearman.'")
        
        return cor
    
    
    @staticmethod
    def residual_variance(x: Optional["np.ndarray"]=None,
                          y: Optional["np.ndarray"]=None,
                          r: Optional[float]=None) -> float:
        '''Residual Variance
        
        The residual variance is computed with the following formuation with r as the
        pearson correlation: 1-r**2. If r is provided, x and y are optional for efficiency.
        
        :param x: The first 1D array, optional. 
        :param y: The second 1D array, optional.
        :param r: Pearson correlation between x and y, optional.

        Returns:
            float: Redisual variance.
        '''
        
        if r is None:
            if x is None or y is None:
                raise ValueError("Either 'r' or both 'x' and 'y' is needed.")
            else:
                cor = EvaluationMetrics.correlation(x, y, metric="Pearson")
                return 1 - cor**2
        elif r > 1 or r < -1:
            raise ValueError("'r' must be between -1 and 1.")
        else:
            return 1 - r**2
        
        
    @staticmethod
    def EMD(x: "np.ndarray",
            y: "np.ndarray",
            normalization: Optional[str] = None) -> float:
        '''Earth Mover's Distance (EMD)
        
        This metric computes the EMD between the pairwise distance of between points in the
        high and low dimensional space. This implementation uses the ``scipy.stats.wasserstein_distance``.
        The usage of EMD is proposed in Heiser & Lou (2020).
        
        :param x: The first distribution x as a 1D array.
        :param y: The second distribution y as a 1D array.
        :param normalization: Whether to perfrom minmax normalization on `x` and `y`. The acceptable
            value is `minmax`, which performs min-max normalization. If `None`, no normalization is performed.
            Defaults to `None`.
            
        :raises ValueError: Unsupported normalization method.
        :return: Earth mover's distance.
        
        .. versionadded:: 0.3.0
        
            The `normalization` parameter. It was added to allow for nromalization of `x` and `y` with
            min-max normalization using `minmax`. This is useful when `x` and `y` are empirical observations
            that are on different scales, and the scale itself is not of interest.
        '''
        if normalization is not None:
            if normalization.lower() == "minmax":
                x = (x-np.min(x))/(np.max(x)-np.min(x))
                y = (y-np.min(y))/(np.max(y)-np.min(y))
            else:
                raise ValueError("Unsupported normalization method. Use 'minmax' or `None`.")
        return scipy.stats.wasserstein_distance(x, y)

    
    @staticmethod
    def KNN(data_neighbors: "np.ndarray",
            embedding_neighbors: "np.ndarray") -> float:
        '''K-Nearest Neighbors Preservation (KNN)
        
        The KNN metric computes the percentage of k-neighbors of each point is preserved in the
        embedding space, and it is average across the entire dataset.
        
        .. note:: This method is not used to calculate KNN itself.

        :param data_neighbors: A nearest-neighbor array of the original data.
        :param embedding_neighbors: A nearest-neighbor array of the embedding.

        :return: K-nearest neighbors preservation.
        '''
        
        i: int
        intersect: int = 0
        for i in range(data_neighbors.shape[0]):
            intersect += np.intersect1d(data_neighbors[i], embedding_neighbors[i]).shape[0]
        
        return intersect/data_neighbors.size

    
    @staticmethod
    def NPE(data_neighbors: "np.ndarray",
            embedding_neighbors: "np.ndarray",
            labels: "np.ndarray",
            method: "str" = "L1") -> float:
        '''Neighborhood Proportion Error (NPE)
        
        The NPE metric is proposed by Konstorum et al. (2019). It measures the total variation distance between
        the proportion of nearest points belonging to the same class of each point in the HD and LD space. The
        lower the NPE, the more similar the embedding and the original data are.
        
        To further elaborate on the difference between L1 norm and TVD, the NPE metric involves the following calculation
        :math:`\\delta (P, Q)`, where :math:`\\delta` is a distance measureon :math:`P` and :math:`Q`. The "L1" optioc
        computes
        
        .. math::
        
            \\sum_i |P_i-Q_i|
            
        whereas the TVD computes
        
        .. math::
        
            \\sup_{a\\in [0,1]} |P(a) - Q(a)| \\, .
            
        There is questionably some debate on the implementation used to calculate NPE, but TVD should align
        more with the original authors' implementation.
        
        :param data_neighbors: A nearest-neighbor array of the original data.
        :param embedding_neighbors: A nearest-neighbor array of the embedding.
        :param labels: The class labels of each observation.
        :param method: The distance measure used for computing the distance between the neighborhood-proportion
            vector. "L1" for L1-norm or "tvd" for Total Variation Distance. The latter is likely the intended
            implementation by Konstorum et al. Defaults to "L1".

        :raises ValueError: Unsupported `method` provided. We only support "L1" or "tvd".
        :return: Neighborhood proportion error.

        .. versionadded:: 0.3.0
        
            The `method` parameter. It was added to allow for alternative implementations. Originally, "L1" was implemented
            and it is stil the default. In this new version, now "tvd" is also an option.
        '''
        
        classes: "np.ndarray"
        classes_index: "np.ndarray"
        classes, classes_index = np.unique(labels, return_inverse=True)
        
        same_class_data: "np.ndarray" = np.zeros(data_neighbors.shape[0])
        same_class_embedding: "np.ndarray" = np.zeros(data_neighbors.shape[0]) 
        
        k: int = data_neighbors.shape[1]
        i: int 
        for  i in range(data_neighbors.shape[0]):
            i_index = classes_index[i]
            i_neighbors_data = data_neighbors[i]
            i_neighbors_embedding = embedding_neighbors[i]
            
            same_class_data[i] = np.count_nonzero(classes_index[i_neighbors_data]==i_index)/k
            same_class_embedding[i] = np.count_nonzero(classes_index[i_neighbors_embedding]==i_index)/k
        
        distance: float = 0.0
        c: str
        if method.lower() == "l1":
            for c in classes:
                P: "np.ndarray" = same_class_data[labels==c]
                Q: "np.ndarray" = same_class_embedding[labels==c]
                distance += np.sum(np.absolute(P-Q))/2
        elif method.lower() == "tvd":
            for c in classes:
                P: "np.ndarray" = same_class_data[labels==c]
                Q: "np.ndarray" = same_class_embedding[labels==c]
                if P.shape[0] < 2 or Q.shape[0] < 2:
                    continue
                try:
                    P_density: "np.ndarray" = scipy.stats.gaussian_kde(P).evaluate(P)
                    Q_density: "np.ndarray" = scipy.stats.gaussian_kde(Q).evaluate(Q)
                except np.linalg.LinAlgError: # pragma: no cover
                    # This is a rare situation of point mass with P or Q
                    continue
                else:
                    distance += np.max(np.absolute(P_density-Q_density))
        else:
            raise ValueError("Method unsupported: Use 'L1' or 'tvd' instead.")
            
        return distance/classes.size
    
    
    @staticmethod
    def neighborhood_agreement(data_neighbors: "np.ndarray",
                               embedding_neighbors: "np.ndarray") -> float:
        '''Neighborhood Agreement
        
        The Neighborhood Agreement metric is proposed by Lee et al. (2015). It measures
        the intersection of k-nearest neighbors (KNN) of each point in HD and LD space. The
        result is subsequently rescaled to measure the improvement over a random embedding.
        This measure is conceptually similar to ``Metric.KNN`` such that they both measure
        the agreement of KNN, but ``Metric.KNN`` simply takes the average of the KNN graph
        agreement without any scaling.
        
        :param data_neighbors: A nearest-neighbor array of the original data.
        :param embedding_neighbors: A nearest-neighbor array of the embedding.

        :return: Neighborhood agreement.
        '''
        
        k: int = data_neighbors.shape[1]
        i: int
        agreement: float = 0.0
        for i in range(data_neighbors.shape[0]):
            agreement += np.intersect1d(data_neighbors[i], embedding_neighbors[i]).shape[0]
        
        agreement = (agreement/(k*data_neighbors.shape[0])*(data_neighbors.shape[0]-1)-k)/(data_neighbors.shape[0]-1-k)
        
        return agreement
    
    
    @staticmethod
    def neighborhood_trustworthiness(data_neighbors: "np.ndarray",
                                     embedding_neighbors: "np.ndarray",
                                     dist_data: "np.ndarray") -> float:
        '''Neighborhood Trustworthiness 
        
        The Neighborhood Truestworthiness is proposed by Venna and Kaski (2001). It measures
        trustworthiness by measuring the ranked distane of new points entering the defined
        neighborhood size in the embedding. The higher the new points are ranked based on the
        original HD space distance matrix, the less trustworthy the new embedding is. The measure
        is scaled between 0 and 1 with a higher score reflecting a more trustworthy embedding.
        
        :param data_neighbors: A nearest-neighbor matrix of the original data.
        :param embedding_neighbors: A nearest-neighbor matrix of the embedding.
        :param dist_data: A pairwise distance matrix for the original data.

        :return: Neighborhood trustworthiness.
        '''
        
        dist_data = scipy.spatial.distance.squareform(dist_data)
        dist_data = scipy.stats.rankdata(dist_data, axis=1)
        
        k: int = data_neighbors.shape[1]
        score: float = 0
        i: int
        for i in range(data_neighbors.shape[0]):
            neighbor_diff: "np.ndarray" = np.setdiff1d(embedding_neighbors[i], data_neighbors[i], assume_unique=True)
            score += np.sum(dist_data[i, neighbor_diff] - k)
            
        score = 1 - 2*score/(data_neighbors.shape[0]*k*(2*data_neighbors.shape[0]-3*k-1))
        
        return score
    
    
    @staticmethod
    def random_forest(embedding: "np.ndarray",
                      labels: "np.ndarray") -> float:
        '''Random Forest Classification Accuracy
        
        This method trains a random forest classifer using the embedding data and the labels
        generated or manually classified from the original space. It then tests the accuracy
        of the classifier using the 33% of the embedding data. This metric was first proposed in
        Becht et al. (2019).
        
        :param embedding: The low-dimensional embedding.
        :param labels: The class labels of each observation.

        :return: Random forest prediction accuracy.
        '''
        
        embedding_train, embedding_test, labels_train, labels_test = train_test_split(embedding, labels, test_size=0.33)

        rf: "RandomForestClassifier" = RandomForestClassifier().fit(embedding_train, labels_train)
        predictions: "np.ndarray" = rf.predict(embedding_test)
        
        return sklearn.metrics.accuracy_score(labels_test, predictions)
    
    
    @staticmethod
    def silhouette(embedding: "np.ndarray",
                   labels: "np.ndarray") -> float:
        '''Silhouette Score
        
        This metric computes the silhouette score of clusters in the embedding space. Ideally,
        clusters should be coherent, and using labels obtained from the original space can
        evaluate the effectiveness of the embedding technique. This metric is used in 
        Xiang et al. (2021).
        
        :param embedding: The low-dimensional embedding.
        :param labels: The class labels of each observation.

        :return: Silhouette score.
        '''
        
        return sklearn.metrics.silhouette_score(embedding, labels)
    
    
    @staticmethod
    def NMI(x_labels: "np.ndarray",
            y_labels: "np.ndarray") -> float:
        '''Normalized Mutual Information (NMI)
        
        The NMI metric computes the mutual information between labels of the original space
        and the embeeding space and then normalizes it with the larger entroy of the two vectors.
        This metric is a measure of clustering performance before and after dimension reduction,
        and it is used in Xiang et al. (2021).
        
        :param x_labels: The first set of labels. 
        :param y_labels: The second set of labels on the same data. 

        :return: Silhouette score.
        '''
        
        mi: float = sklearn.metrics.mutual_info_score(x_labels, y_labels)
        
        labels_count: "np.ndarray"
        embedding_count: "np.ndarray"
        
        _ , labels_count = np.unique(x_labels, return_counts=True)
        _ , embedding_count = np.unique(y_labels, return_counts=True)
        
        labels_count = labels_count/x_labels.size
        embedding_count = embedding_count/y_labels.size
        normalization: float = np.maximum(scipy.stats.entropy(labels_count),
                                          scipy.stats.entropy(embedding_count))
        
        return mi/normalization
    
    
    @staticmethod
    def ARI(x_labels: "np.ndarray",
            y_labels: "np.ndarray") -> float:
        '''Adjusted Rand Index (ARI)
        
        The ARI uses the labels from the original space and the embedding space to measure
        the similarity between them using pairs. It is used in Xiang et al. (2021).
        
        :param x_labels: The first set of labels. 
        :param y_labels: The second set of labels on the same data. 

        :return: ARI.
            
        :References:
        
        This implementation adapts from sklearn's implementation of ARI with a bug fix of overflow
        issue. 
        
        .. code-block:: text
        
            @article{scikit-learn,
            title={Scikit-learn: Machine Learning in {P}ython},
            author={Pedregosa, F. and Varoquaux, G. and Gramfort, A. and Michel, V.
                    and Thirion, B. and Grisel, O. and Blondel, M. and Prettenhofer, P.
                    and Weiss, R. and Dubourg, V. and Vanderplas, J. and Passos, A. and
                    Cournapeau, D. and Brucher, M. and Perrot, M. and Duchesnay, E.},
            journal={Journal of Machine Learning Research},
            volume={12},
            pages={2825--2830},
            year={2011}
            }
            
        :License:
        
        .. code-block:: text
        
            BSD 3-Clause License

            Copyright (c) 2007-2021 The scikit-learn developers.
            All rights reserved.

            Redistribution and use in source and binary forms, with or without
            modification, are permitted provided that the following conditions are met:

            * Redistributions of source code must retain the above copyright notice, this
            list of conditions and the following disclaimer.

            * Redistributions in binary form must reproduce the above copyright notice,
            this list of conditions and the following disclaimer in the documentation
            and/or other materials provided with the distribution.

            * Neither the name of the copyright holder nor the names of its
            contributors may be used to endorse or promote products derived from
            this software without specific prior written permission.

            THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
            AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
            IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
            DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
            FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
            DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
            SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
            CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
            OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
            OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
        '''
        
        confusion: "np.ndarray" = sklearn.metrics.pair_confusion_matrix(x_labels, y_labels)
        
        tn: int = int(confusion[0][0])
        fp: int = int(confusion[0][1])
        fn: int = int(confusion[1][0])
        tp: int = int(confusion[1][1])
        
        return 2. * (tp * tn - fn * fp) / ((tp + fn) * (fn + tn) + (tp + fp) * (fp + tn))
    
    
    @staticmethod
    def calinski_harabasz(embedding: "np.ndarray",
                          labels: "np.ndarray") -> float:
        '''Calinski-Harabasz Index
        
        This metric computes the Calinski-Harabasz index of clusters in the embedding space. Ideally,
        clusters should be coherent, and using labels obtained from the original space can
        evaluate the effectiveness of the embedding technique.
        
        :param embedding: The low-dimensional embedding.
        :param labels: The class labels of each observation.

        :return: Calinski-Harabasz Index.
        '''
        
        return sklearn.metrics.calinski_harabasz_score(embedding, labels)
    
    
    @staticmethod
    def davies_bouldin(embedding: "np.ndarray",
                       labels: "np.ndarray") -> float:
        '''Davies-Bouldin Index
        
        This metric computes the Davies-Bouldin index of clusters in the embedding space. Ideally,
        clusters should be coherent, and using labels obtained from the original space can
        evaluate the effectiveness of the embedding technique.
        
        :param embedding: The low-dimensional embedding.
        :param labels: The class labels of each observation.

        :return: Davies-Bouldin Index.
        '''
        
        return sklearn.metrics.davies_bouldin_score(embedding, labels)
    
    
    @staticmethod
    def embedding_concordance(embedding: "np.ndarray",
                              labels_embedding: "np.ndarray",
                              comparison_file: Union["np.ndarray", List["np.ndarray"]],
                              comparison_labels: Union["np.ndarray", List["np.ndarray"]],
                              comparison_classes: Optional[List[str]]=None,
                              method: str = "emd"
                              ) -> Union[float, str]:
        """Concordance between two embeddings.
        
        This is a wrapper function to implement two embedding concordance metrics based on
        named clusters: EMD and Cluster Distance. When two embeddings can be reasonably
        aligned based on clusters or manual labels, these two metrics calculate the relationships
        between clusters and their distances between two embeddings.
        
        For EMD, the metric considers matched pairs of clusters in both embeddings: for each pair
        in each embedding, the distances between each centroid and all points in the other cluster
        are calculated. The EMD between these two vectors from two embeddings are calculated and
        then averaged across all pairs. 
        
        For Cluster Distance, pairwise rank distance between all cluster centroids are calculated
        in each embedding. Then, the Euclidean distance between these two vectors are taken.
        
        :param embedding: The first (main) embedding.
        :param labels_embedding: Labels for all obervations in the embedding.
        :param comparison_file: The second embedding.
        :param comparison_labels: The labels for all observations in the comparison embedding.
        :param comparison_classes: Which classes in labels to compare. At least two classes need
            to be provided for this to work; otherwise, `NA` will be returned. If ``None``, all
            overlapping labels used, optional
        :param method: "emd" or "cluster_distance", defaults to "emd"
            
        :return: The score or "NA"
            
        .. Note:: When there is no overlapping labels, "NA" is automatically returned as ``str``.
        
        .. deprecated:: 0.2.0
        
            Passing in `str` for the `comparison_classes` parameter is deprecated
            and will be removed in futrue versions.
        """
        
        if not isinstance(comparison_file, list):
            comparison_file = [comparison_file]
        if not isinstance(comparison_labels, list):
            comparison_labels = [comparison_labels]
        if not isinstance(comparison_classes, list) and comparison_classes is not None:
            comparison_classes = [comparison_classes]
            warnings.warn("Passing in a non-list parameter is deprecated. Use a list instead.", DeprecationWarning, stacklevel=2)
            
        method = method.lower()
            
        scores: "np.ndarray" = np.zeros(len(comparison_file))
        i: int
        comparison_labels_index: int
        comparison_status: bool = False
        
        for i in range(len(comparison_file)):
            comparison_labels_index = i if len(comparison_labels) == len(comparison_file) else 0
            common_types: "np.ndarray" = np.intersect1d(np.unique(labels_embedding), np.unique(comparison_labels[comparison_labels_index]))
            
            if comparison_classes is not None:
                common_types = np.intersect1d(common_types, comparison_classes)     
            if common_types.shape[0] < 2:
                continue
            
            if method == "emd":
                scores[i] = EvaluationMetrics._concordance_emd(embedding,
                                                               labels_embedding,
                                                               comparison_file[i],
                                                               comparison_labels[comparison_labels_index],
                                                               common_types)
            elif method == "cluster_distance":
                scores[i] = EvaluationMetrics._concordance_cluster_distance(embedding,
                                                                            labels_embedding,
                                                                            comparison_file[i],
                                                                            comparison_labels[comparison_labels_index],
                                                                            common_types)
            comparison_status = True
            
        if comparison_status:
            return np.mean(scores)
        else:
            return "NA"
            
            
    @staticmethod
    def _concordance_emd(embedding: "np.ndarray",
                         labels_embedding: "np.ndarray",
                         comparison_file: "np.ndarray",
                         comparison_labels: "np.ndarray",
                         common_types: "np.ndarray") -> float:
        
        """Embedding concordance EMD.
        
        This is a private method that computes the concordance EMD as described in the ``embedding_concordance`` method.
        
        :param embedding: The first (main) embedding.
        :param labels_embedding: Labels for all obervations in the embedding.
        :param comparison_file: The second embedding.
        :param comparison_labels: The labels for all observations in the comparison embedding.
        :param common_types: The common cluster labels for EMD computation.
            
        :return: The EMD score
        """
        
        combinations: List[Tuple[int, ...]] = list(itertools.permutations(common_types, 2))
        embedding_scores: List[float] = []
        comparison_scores: List[float] = []
        
        comb: Tuple[int, ...]
        for comb in combinations:
            indicies_embedding: "np.ndarray" = np.where(labels_embedding == comb[0])[0]
            centroid: "np.ndarray" = np.mean(embedding[indicies_embedding, ], axis=0)
            pwd: "np.ndarray" = scipy.stats.rankdata(np.sqrt(np.sum((centroid - embedding)**2, axis=1)))
            pwd = pwd[np.where(labels_embedding == comb[1])]/(embedding.shape[0])
            embedding_scores.extend(d for d in pwd)
            
            indicies_comparison: "np.ndarray" = np.where(comparison_labels == comb[0])[0]
            centroid_c: "np.ndarray" = np.mean(comparison_file[indicies_comparison,], axis=0)
            pwd_c: "np.ndarray" = scipy.stats.rankdata(np.sqrt(np.sum((centroid_c - comparison_file)**2, axis=1)))
            pwd_c = pwd_c[np.where(comparison_labels == comb[1])]/(comparison_file.shape[0])
            comparison_scores.extend([d for d in pwd_c])
            
        return scipy.stats.wasserstein_distance(embedding_scores, comparison_scores)
    
    
    @staticmethod
    def _concordance_cluster_distance(embedding: "np.ndarray",
                                      labels_embedding: "np.ndarray",
                                      comparison_file: "np.ndarray",
                                      comparison_labels: "np.ndarray",
                                      common_types: "np.ndarray") -> float:
        """Embedding concordance Cluster Distance.
        
        This is a private method that computes the concordance Ckuster Distance metric as described in the
        ``embedding_concordance`` method.
        
        :param embedding: The first (main) embedding.
        :param labels_embedding: Labels for all obervations in the embedding.
        :param comparison_file: The second embedding.
        :param comparison_labels: The labels for all observations in the comparison embedding.
        :param common_types: The common cluster labels for EMD computation.
            
        :return: The EMD score
        """
        
        combinations: List[Tuple[int, ...]] = list(itertools.permutations(common_types, 2))
        embedding_scores: "np.ndarray" = np.zeros(len(combinations))
        comparison_scores: "np.ndarray" = np.zeros(len(combinations))
        
        embedding_clusters: "np.ndarray" = np.unique(labels_embedding)
        embedding_centroid: "np.ndarray" = np.zeros((embedding_clusters.shape[0], embedding.shape[1]))
        for i, c in enumerate(embedding_clusters):
            embedding_centroid[i] = np.mean(embedding[np.where(labels_embedding==c)[0],], axis=0)
                
        comparison_clusters: "np.ndarray" = np.unique(comparison_labels)
        comparison_centroid: "np.ndarray" = np.zeros((comparison_clusters.shape[0], comparison_file.shape[1]))
        for i, c in enumerate(comparison_clusters):
            comparison_centroid[i] = np.mean(comparison_file[np.where(comparison_labels==c)[0],], axis=0)

        comb: Tuple[int, ...]
        i: int
        for i, comb in enumerate(combinations):
            embedding_index_0: int = np.where(embedding_clusters == comb[0])[0][0]
            embedding_index_1: int = np.where(embedding_clusters == comb[1])[0][0]
            embedding_cluster_0: "np.ndarray" = embedding_centroid[embedding_index_0]
            embedding_pwd: "np.ndarray" = scipy.stats.rankdata(np.sqrt(np.sum((embedding_cluster_0 - embedding_centroid)**2, axis=1)))
            embedding_scores[i] = embedding_pwd[embedding_index_1,]/embedding_clusters.shape[0]
            
            comparison_index_0: int = np.where(comparison_clusters == comb[0])[0][0]
            comparison_index_1: int = np.where(comparison_clusters == comb[1])[0][0]
            comparison_cluster_0: "np.ndarray" = comparison_centroid[comparison_index_0]
            comparison_pwd: "np.ndarray" = scipy.stats.rankdata(np.sqrt(np.sum((comparison_cluster_0 - comparison_centroid)**2, axis=1)))
            comparison_scores[i] = comparison_pwd[comparison_index_1,]/comparison_clusters.shape[0]
            
        return np.mean(np.abs(embedding_scores-comparison_scores))  
    
    
class PointClusterDistance():
    """Point Cluster Distance
    
    This class is used to compute the Point Cluster Distance. Instead of full pairwise
    distance, this distance metric computes the distance between each cluster centroid
    and all other point. The memory complexity is N_cluster*N instead of (N^2)/2.
    
    :param X: The input data array.
    :param labels: Labels for the data array.
    :param dist_metric: The distance metric to use. This supports "euclidean", "manhattan", or "cosine", defaults to "euclidean"
        
    :Attributes:
        - **X**: The input data array.
        - **labels**: Labels for the data array.
        - **dist_metric**: The distance metric to use. This supports "euclidean", "manhattan", or "cosine", defaults to "euclidean"
        - **dist**: The calculated distance array. The first axis corresponds to each observation in the original array and the second axis is all the cluster centroids, optional.
    """
    
    def __init__(self, X: "np.ndarray", labels: "np.ndarray", dist_metric: str="euclidean"):
        self.X: "np.ndarray" = X
        self.labels: "np.ndarray" = labels
        self.dist_metric = dist_metric.lower()
        self.dist: Optional["np.ndarray"] = None
        self.index: Optional["np.ndarray"] = None
        
        
    def fit(self, flatten: bool=False) -> "np.ndarray":
        """Fit the distance metric.

        This method calculates the distance metric based on the class attributes.
        
        :param flatten: Whether to flatten the return into a 1-d vector

        :return: The calculate distance array.
        """
        index: "np.ndarray"
        inverse: "np.ndarray"
        index, inverse = np.unique(self.labels, return_inverse=True)
        
        self.dist = np.empty((self.X.shape[0], index.size))
        self.index = index
        centroid: "np.ndarray" = self._cluster_centroid(self.X, index, inverse)
        dist = self._distance_func(dist_metric=self.dist_metric)
        
        i: int
        obs: "np.ndarray"
        for i, obs in enumerate(self.X):
            self.dist[i] = dist(obs, centroid)
            
        if flatten:
            return self.flatten(self.dist)
        else:
            return self.dist
        
    
    @staticmethod
    def flatten(dist: "np.ndarray") -> np.ndarray:
        """Flatten an array

        This method is a wrapper for the ``flatten`` method in ``numpy``.

        :param dist: The distance array.

        :return: The flattened array.
        """
        return dist.flatten()
    
    
    @staticmethod
    def _distance_func(dist_metric: str="euclidean") -> Callable[[np.ndarray, np.ndarray], float]:
        if dist_metric == "manhattan":
            return PointClusterDistance._manhattan
        elif dist_metric == "cosine":
            return PointClusterDistance._cosine
        else:    
            return PointClusterDistance._euclidean
    
    
    @staticmethod
    def _euclidean(X1: np.ndarray, X2: np.ndarray) -> float:
        """Euclidean distance

        This is an implementation of the Euclidean distance.

        Args:
        :param X1: The array to calculate Euclidean distance.
        :param X2: The array to calculate Euclidean distance.

        :return: The euclidean distance
        """
        return np.sqrt(np.sum(np.square(X1-X2), axis=1))
    
    
    @staticmethod
    def _manhattan(X1: np.ndarray, X2: np.ndarray) -> float:
        """Euclidean distance

        This is an implementation of the Manhattan distance.

        Args:
        :param X1: The array to calculate Manhattan distance.
        :param X2: The array to calculate Manhattan distance.

        :return: The Manhattan distance
        """
        return np.sum(np.abs(X1-X2), axis=1)
    
    
    @staticmethod
    def _cosine(X1: np.ndarray, X2: np.ndarray) -> float:
        """Cosine distance

        This is an implementation of the Cosine distance.

        Args:
        :param X1: The array to calculate Cosine distance.
        :param X2: The array to calculate Cosine distance.

        :return: The Cosine distance
        """
        return np.sum(X1*X2, axis=1)/(np.sqrt(np.sum(X1**2))*np.sqrt(np.sum(X2**2, axis=1)))
    
    
    @staticmethod
    def _cluster_centroid(X: np.ndarray, index: np.ndarray, inverse: np.ndarray) -> np.ndarray:
        """Find the centroids of clustered data.

        With the input array and labels, this method computes all centroids using
        the labels as clusters.

        :param X: The original data array.
        :param index: The unique IDs of the observations.
        :param inverse: The indicies to reconstruct the original array.

        :return: The cluster centroids.
        """
        centroid: "np.ndarray" = np.empty((index.size, X.shape[1]))
        for i in range(len(index)):
            centroid[i] = np.mean(X[inverse==i])
            
        return centroid


class Annoy():
    
    @staticmethod
    def build_annoy(data: "np.ndarray",
                    metric: str = "angular",
                    n_trees: int=10) -> "AnnoyIndex":
        """Build ``AnnoyIndex`` object from data.

        :param data: The data array
        :param metric: The distance metric to use, defaults to "angular"
        :param n_trees: The number of trees, defaults to 10
        
        :return: An ``AnnoyIndex`` object.
        """
        
        model: "AnnoyIndex" = AnnoyIndex(data.shape[1], metric = metric) #type: ignore
        for i in range(data.shape[0]):
            model.add_item(i, data[i])
        model.build(n_trees=n_trees, n_jobs=-1)
        return model
    
    
    @staticmethod
    def load_annoy(path: str,
                   ncol: int,
                   metric: str = "angular"
                   ) -> "AnnoyIndex":
        """Load ``AnnoyIndex`` object from disk.
        
        This loads an AnnoyIndex object saved using this class or the ANnoy's buildin IO function.   

        :param path: The path to the object.
        :param ncol: The number of columns.
        :param metric: _description_, defaults to "angular"
        :return: The loaded ``AnnoyIndex`` object.
        """
        model: "AnnoyIndex" = AnnoyIndex(ncol, metric) #type: ignore
        model.load(path)
        return model
    
    
    @staticmethod
    def save_annoy(model: "AnnoyIndex", path: str):
        """Save ``AnnoyIndex`` object to disk.
        
        This saves an ``AnnoyIndex`` object to a specified path.   

        :param model: An ``AnnoyIndex`` object to be saved.
        :param path: The path to the object.
        :return: The loaded ``AnnoyIndex`` object.
        """
        model.save(path)


import setuptools
import sys
import os
import shutil
import distutils.cmd

VERSION = "0.3.0"

class PypiCommand(distutils.cmd.Command):
    
    description = "Build and upload for PyPI."
    user_options = []
    
    def initialize_options(self):
        pass
    
    
    def finalize_options(self):
        pass
    
    
    def run(self):
        shutil.rmtree("dist/")
        
        wheel_file = "CytofDR-{}-py3-none-any.whl".format(VERSION)
        tar_file = "CytofDR-{}.tar.gz".format(VERSION)
        
        os.system("{} setup.py sdist bdist_wheel".format(sys.executable))
        os.system("twine upload dist/{} dist/{}".format(wheel_file, tar_file))
    
    
class CondaCommand(distutils.cmd.Command):
    
    description = "Build and upload for conda."
    user_options = []
    
    def initialize_options(self):
        pass
    
    
    def finalize_options(self):
        pass
    
    
    def run(self):
        shutil.rmtree("dist_conda/")
        os.system("conda build . --output-folder dist_conda/ -c conda-forge -c bioconda")
        os.system("anaconda upload ./dist_conda/noarch/cytofdr-{}-py_0.tar.bz2".format(VERSION))


setuptools.setup(
    name = "CytofDR",
    version = VERSION,
    description = "CyTOF Dimension Reduction Framework",
    author="Kevin Wang",
    url="https://github.com/kevin931/CytofDR",
    long_description_content_type = "text/markdown",
    long_description = open("README.md").read(),
    packages=["CytofDR"],
    python_requires=">=3.7",
    install_requires=["scikit-learn",
                      "numpy",
                      "scipy",
                      "umap-learn",
                      "openTSNE",
                      "phate",
                      "annoy",
                      "matplotlib",
                      "seaborn"],
    test_requires=["pytest",
                   "pytest-cov",
                   "pytest-mock",
                   "coverage"],
    license="MIT",
    classifiers = [
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3 :: Only",
        "Natural Language :: English"
    ],
    cmdclass = {"pypi": PypiCommand,
                "conda": CondaCommand
                }
)


import numpy as np
import scipy
from sklearn.decomposition import FastICA, PCA, FactorAnalysis, KernelPCA, NMF
from sklearn.manifold import Isomap, MDS, LocallyLinearEmbedding, SpectralEmbedding, TSNE
from sklearn.cluster import KMeans

import umap
from openTSNE import TSNEEmbedding
from openTSNE import affinity
from openTSNE import initialization
import phate

import scipy.spatial
import scipy.stats
from CytofDR.evaluation import EvaluationMetrics, PointClusterDistance

import seaborn as sns
import matplotlib.pyplot as plt

import _csv
import csv
import os
import warnings
from typing import Union, Optional, List, Dict, Any, Set

METHODS: Dict[str, bool]  = {"SAUCIE": True, "ZIFA": True, "GrandPrix": True}
    
try:
    import SAUCIE
except ImportError:
    METHODS["SAUCIE"] = False
    print("No 'saucie' implementation.")
    
try:
    from ZIFA import ZIFA
except ImportError:
    METHODS["ZIFA"] = False
    print("No 'ZIFA' implementation.")
     
try:
    from GrandPrix import GrandPrix
except ImportError:
    METHODS["GrandPrix"] = False
    print("No 'Grandprix' implementation.")
    
    
def _verbose(message: str, verbose:bool=True):
    if verbose:
        print(message)
    
  
class Reductions():
    """A class for reductions and their evaluation. 

    This class is a convenient data class for storing and evaluaqting reductions.

    :param reductions: A dictionary of reductions as indexed by their names.
    
    :Attributes:
        - **reductions**: A dictionary of reductions as indexed by names.
        - **names**: The names of the reductions.
        - **original_data**: The original space data before DR.
        - **original_labels**: Clusterings based on original space data.
        - **original_cell_types**: Cell types based on original space data. 
        - **embedding_data**: The embedding space reduction.
        - **embedding_labels**: Clusterings based on embedding space reduction.
        - **embedding_cell_types**: Cell types based on embedding space reduction. 
        - **comparison_data**: The comparison data (matched with original data in some way) for concordance analysis.
        - **comparison_cell_types**: Cell types based on comparison data.
        - **comparison_classes**: Common cell types between embedding and comparison data.
        - **evaluations**: The DR evaluation results, which are generated by the default evaluation method.
        - **custom_evaluations**: The custom DR evaluation results, which are added by the ``add_custom_evaluation_results()`` method.
        - **custom_evaluation_weights**: The weights for each metric in the ``custom_evaluations``. They do not distinguish between same metrics on different reductions.
        - **custom_evaluation_reverse_ranking**: Whether each metric in the ``custom_evaluations`` should be reverse ranked (i.e. smaller is better).
    """
    
    def __init__(self, reductions: Optional[Dict[str, "np.ndarray"]]=None):
        """Constructor method for Reductions.
        """
        self._reductions: Dict[str, "np.ndarray"]
        self.names: Set[str]
        if reductions is None:
            self._reductions = {}
            self.names = set()
        else:
            self._reductions = reductions
            self.names = set(self._reductions.keys())
        self.original_data: Optional["np.ndarray"] = None
        self.original_labels: Optional["np.ndarray"] = None
        self.original_cell_types: Optional["np.ndarray"] = None
        self.embedding_labels: Optional[Dict[str, "np.ndarray"]] = None
        self.embedding_cell_types: Optional[Dict[str, "np.ndarray"]] = None
        self.comparison_data: Optional["np.ndarray"] = None
        self.comparison_cell_types: Optional["np.ndarray"] = None
        self.comparison_classes: Optional[Union[str, List[str]]] = None
        
        self.evaluations: Dict[str, Any] = {}
        self.custom_evaluations: Dict[str, Any] = {"custom": {}}
        self.custom_evaluation_weights: Dict[str, Any] = {}
        self.custom_evaluation_reverse_ranking: Dict[str, Any] = {}
    
    
    def add_reduction(self, reduction: "np.ndarray", name: str, replace: bool=False):
        """Add a reduction embedding.

        This method allows users to add additional embeddings.

        :param reduction: The reduction array.
        :param name: The name of the reduction.
        :param replace: If the original name exists, whether to replace the original, defaults to False
        
        :raises ValueError: Reduction already exists but users choose not to replace the original.
        """
        if name in self.names and not replace:
            raise ValueError("Reduction already exists. Set 'replace' to True if replacement is intended.")
        self.reductions[name] = reduction
        self.names.add(name)
        
        
    def get_reduction(self, name: str):
        """Retrieve a reduction my name.

        This method allows users to retrieve a reduction by name. This equivalent to running ``self.reductions[name]``.

        :param name: The name of the reduction.
        """
        return self.reductions[name]
    
    
    def add_evaluation_metadata(self,
                                original_data: Optional["np.ndarray"]=None,
                                original_labels: Optional["np.ndarray"]=None,
                                original_cell_types: Optional["np.ndarray"]=None,
                                embedding_labels: Optional[Dict[str, "np.ndarray"]]=None,
                                embedding_cell_types: Optional[Dict[str, "np.ndarray"]]=None,
                                comparison_data: Optional["np.ndarray"]=None,
                                comparison_cell_types: Optional["np.ndarray"]=None,
                                comparison_classes: Optional[Union[str, List[str]]]=None,
                                ):
        """Add supporting metadata for DR evaluation.

        This method allows you to add metadata in the process of DR evaluation. They do not
        override existing metadata unless actual inputs are specified.
        
        :param original_data: The original space data before DR.
        :param original_labels: Clusterings based on original space data.
        :param original_cell_types: Cell types based on original space data. 
        :param embedding_data: The embedding space reduction.
        :param embedding_labels: Clusterings based on embedding space reduction.
        :param embedding_cell_types: Cell types based on embedding space reduction. 
        :param comparison_data: The comparison data (matched with original data in some way) for concordance analysis.
        :param comparison_cell_types: Cell types based on comparison data.
        :param comparison_classes: Common cell types between embedding and comparison data.
        """
        if original_data is not None:
            self.original_data = original_data
        if original_labels is not None:
            self.original_labels = original_labels
        if original_cell_types is not None:
            self.original_cell_types = original_cell_types
        if embedding_labels is not None:
            self.embedding_labels = embedding_labels
        if embedding_cell_types is not None:
            self.embedding_cell_types = embedding_cell_types
        if comparison_data is not None:
            self.comparison_data = comparison_data
        if comparison_cell_types is not None:
            self.comparison_cell_types = comparison_cell_types
        if comparison_classes is not None:
            self.comparison_classes = comparison_classes
            
            
    def cluster(self, n_clusters: int, cluster_data: bool = True, cluster_embedding: bool = True, **kwargs):
        """Cluster original_data and reductions.

        This provides a convenient method to cluster `original_data` and embeddings using the
        ``KMeans`` method. The number of clusters must be manually specified.

        :param n_clusters: The number of clusters
        :param cluster_data: Whether to cluster original data, defaults to True
        :param cluster_embedding: Whether to cluster embeddings, defaults to True
        :param kwargs: Keyword only arguments passed into ``sklearn.cluster.KMeans``.
        """
        
        if cluster_data and self.original_data is None:
            raise ValueError("'original_data' is missing: cannot cluster.")
        
        if cluster_embedding and len(self.reductions) == 0:
            raise ValueError("'reductions' is missing: cannot cluster.")
        
        if cluster_data:
            self.original_labels = KMeans(n_clusters=n_clusters, **kwargs).fit(self.original_data).labels_
        
        if cluster_embedding:
            self.embedding_labels = {}
            for e in self.reductions.keys():
                self.embedding_labels[e] = KMeans(n_clusters=n_clusters, **kwargs).fit(self.reductions[e]).labels_
        
    
    def evaluate(self,
                 category: Union[str, List[str]],
                 pwd_metric: str="PCD",
                 k_neighbors: int=5,
                 annoy_original_data_path: Optional[str]=None,
                 auto_cluster: bool=True,
                 n_clusters: int=20,
                 verbose: bool=True,
                 pairwise_downsample_size: int=10000,
                 normalize_pwd: Optional[str]=None,
                 NPE_method: str="L1"):     
        """Evaluate DR Methods Using Default DR Evaluation Scheme.

        This method ranks the DR methods based on any of the four default categories:
        ``global``, ``local``, ``downstream``, or ``concordance``. 
        
        :param category: The major evaluation category: ``global``, ``local``, ``downstream``, or ``concordance``.
        :param pwd_metric: The pairwise distance metric. Three options are "PCD", "pairwise_downsample", or "pairwise".
            PCD refers to Point Cluster Distance as implemented in this package; pairwise 
            is the traditional pairwise distance; pairwise_downsample is pairwise distance with downsampling
            for both data and embedding. For large datasets, PCD and pairwise_downsample are recommended and
            practically with equivalent performances. Defaults to "PCD".
        :param k_neighbors: The number of neighbors to use for ``local`` metrics. Defaults to 5.
        :param annoy_original_data_path: The file path to an ANNOY object for original data. Optional.
        :param auto_cluster: Whether to automatically perform clustering for evaluation purposes.
            This option has no effect when the `original_labels` and `embedding_labels` are previously added
            with the `add_evaluation_metadata` method. Defaults to `True`.
        :param n_clusters: The number of clusters for the `auto_cluster` option. Defaults to 20.
        :param verbose: Whether to print out progress. Defaults to `True`.
        :param pairwise_downsample_size: The downsample size if the `pairwise_downsample` is chosen for the
            `pwd_metric`. If this is larger than the sample size of the original dataset, this methods falls
            back to the `pairwise` option for `pwd_metric`. On a typical machine, it is not recommended to 
            go beyond the default, and for datasets smaller than 10,000, no downsample is strictly necessary.
            Defaults to 10,000.
        :param normalize_pwd: Whether to perform minmax normalize on the pwd metric for EMD. If needed, enter
            "minmax". If ``None``, then the raw distances are used. For more details, see ``EvaluationMetrics.EMD``.
            This only matters when ``Global`` is chosen in the ``category`` parameter. Defaults to None.
        :param NPE_method: The distance measure used for the NPE metric as part of the ``Local`` categorty.
            "L1" for L1-norm or "tvd" for Total Variation Distance. Defaults to "L1". 
        
        :raises ValueError: No reductions to evalate.
        :raises ValueError: Unsupported 'pwd_metric': 'PCD', 'Pairwise', or 'pairwise_downsample' only.
        :raises ValueError: Evaluation needs 'original_data', 'original_labels', and 'embedding_labels' attributes.
        
        .. note::
        
            This method requires ``add_evaluation_metadata`` to run first. ``original_cell_types`` and
            ``embedding_cell_types`` are optional for the downstream category. For ``concordance``, if
            you wish to use clustering results for embedding and comparison files, set the appropriate
            clusterings to ``embedding_cell_types`` and ``comparison_cell_types``.
            
        .. versionadded:: 0.2.0 
        
            The `pairwise_downsample` option for `pwd_metric`; the `pairwise_downsample_size` parameter for when
            `pairwise_downbsample` is chosen.
            
        .. versionadded:: 0.3.0
        
            The `normalize_pwd` parameter. It was added to allow for nromalization in the EMD metric for ``Global``.
            This should be used in cases that there are scale differences in different DR embeddings. For this
            specific update, only "minmax" is supported.
            
        .. versionadded:: 0.3.0
        
            The `NPE_method` parameter. It was added to allow for alternative implementations of NPE's distance measure.
            Originally, "L1" was implemented and it is stil the default. In this new version, now "tvd" is also an option.
            Defaults to "L1".
        """
        
        if len(self.reductions) == 0:
            raise ValueError("No reductions to evalate. Add your reductions first.")
        if pwd_metric.lower() not in  ["pcd", "pairwise", "pairwise_downsample"]:
            raise ValueError("Unsupported 'pwd_metric': 'PCD', 'Pairwise', or 'pairwise_downsample' only.")
        
        self.evaluations = {}
                
        if isinstance(category, list):
            category = [c.lower() for c in category]
        else:
            category = [category.lower()]
            
        if self.original_labels is None and auto_cluster:
            self.cluster(n_clusters=n_clusters, cluster_embedding=False)
        if self.embedding_labels is None and auto_cluster:
            self.cluster(n_clusters=n_clusters, cluster_data=False)
                                
        if self.original_labels is None or self.embedding_labels is None:
            message: str = "Evaluation needs 'original_labels', and 'embedding_labels' attributes. "
            message += "Use the 'auto_cluster' option or add your own lables with the 'add_evaluation_metadata()' method."
            raise ValueError(message)
        
        if self.original_data is None:
            raise ValueError("Evaluation needs 'original_data'. Please add it with the 'add_evaluation_metadata()' method.")
        
        e: str
        if "global" in category:
            _verbose("Evaluating global...", verbose=verbose)
            self.evaluations["global"] = {"spearman": {}, "emd": {}}
            data_distance: "np.ndarray"
            embedding_distance: Dict[str, "np.ndarray"] = {}
            
            if pwd_metric.lower() == "pcd":
                data_distance = PointClusterDistance(self.original_data, self.original_labels).fit(flatten=True)
                for e in self.reductions.keys():
                    embedding_distance[e] = PointClusterDistance(self.reductions[e], self.original_labels).fit(flatten=True)
                    
            elif pwd_metric.lower() == "pairwise_downsample":
                downsample_index: np.ndarray
                if pairwise_downsample_size >= self.original_data.shape[0]:
                    downsample_index = np.arange(self.original_data.shape[0])
                else:
                    downsample_index = np.random.choice(np.arange(self.original_data.shape[0]), size=pairwise_downsample_size, replace=False)
                data_distance = scipy.spatial.distance.pdist(self.original_data[downsample_index])
                for e in self.reductions.keys():
                    embedding_distance[e] = scipy.spatial.distance.pdist(self.reductions[e][downsample_index], metric="euclidean")
                
            else:
                data_distance = scipy.spatial.distance.pdist(self.original_data)
                for e in self.reductions.keys():
                    embedding_distance[e] = scipy.spatial.distance.pdist(self.reductions[e], metric="euclidean")
                    
            for e in self.reductions.keys():
                val: float = EvaluationMetrics.correlation(x=data_distance, y=embedding_distance[e], metric="Spearman")
                self.evaluations["global"]["spearman"][e] = val
                
                val: float = EvaluationMetrics.EMD(x=data_distance, y=embedding_distance[e], normalization=normalize_pwd)
                self.evaluations["global"]["emd"][e] = val
                   
        if "local" in category:
            _verbose("Evaluating local...", verbose=verbose)
            assert self.original_labels is not None
            self.evaluations["local"] = {"knn": {}, "npe": {}}
            
            data_neighbors: "np.ndarray" = EvaluationMetrics.build_annoy(self.original_data, annoy_original_data_path, k_neighbors)
            for e in self.reductions.keys():
                embedding_neighbors: "np.ndarray" = EvaluationMetrics.build_annoy(self.reductions[e], None, k_neighbors)
                self.evaluations["local"]["npe"][e] = EvaluationMetrics.NPE(labels = self.original_labels, data_neighbors=data_neighbors, embedding_neighbors=embedding_neighbors, method=NPE_method)
                self.evaluations["local"]["knn"][e] = EvaluationMetrics.KNN(data_neighbors=data_neighbors, embedding_neighbors=embedding_neighbors)
                             
        if "downstream" in category:
            _verbose("Evaluating downstream...", verbose=verbose)
            self.evaluations["downstream"] = {"cluster reconstruction: silhouette": {},
                                              "cluster reconstruction: DBI": {},
                                              "cluster reconstruction: CHI": {},
                                              "cluster reconstruction: RF": {},
                                              "cluster concordance: ARI": {},
                                              "cluster concordance: NMI": {},
                                              "cell type-clustering concordance: ARI": {},
                                              "cell type-clustering concordance: NMI": {}}
            for e in self.reductions.keys():
                self.evaluations["downstream"]["cluster reconstruction: silhouette"][e] = EvaluationMetrics.silhouette(embedding=self.reductions[e],
                                                                                                             labels=self.original_labels)
                self.evaluations["downstream"]["cluster reconstruction: DBI"][e] = EvaluationMetrics.davies_bouldin(embedding=self.reductions[e],
                                                                                                                    labels=self.original_labels)
                self.evaluations["downstream"]["cluster reconstruction: CHI"][e] = EvaluationMetrics.calinski_harabasz(embedding=self.reductions[e],
                                                                                                                       labels=self.original_labels)
                self.evaluations["downstream"]["cluster reconstruction: RF"][e] = EvaluationMetrics.random_forest(embedding=self.reductions[e],
                                                                                                                 labels=self.original_labels)
                
                self.evaluations["downstream"]["cluster concordance: ARI"][e] = EvaluationMetrics.ARI(x_labels=self.original_labels,
                                                                                                      y_labels=self.embedding_labels[e])
                self.evaluations["downstream"]["cluster concordance: NMI"][e] = EvaluationMetrics.NMI(x_labels=self.original_labels,
                                                                                                      y_labels=self.embedding_labels[e])
                
                if self.original_cell_types is not None:
                    self.evaluations["downstream"]["cell type-clustering concordance: ARI"][e] = EvaluationMetrics.ARI(x_labels=self.original_cell_types,
                                                                                                                    y_labels=self.embedding_labels[e])
                    self.evaluations["downstream"]["cell type-clustering concordance: NMI"][e] = EvaluationMetrics.NMI(x_labels=self.original_cell_types,
                                                                                                                    y_labels=self.embedding_labels[e])
                else:
                    warnings.warn("No 'original_sell_types': Cell type-clustering concordance is not evaluated.")
            
        if "concordance" in category:
            _verbose("Evaluating concordance...", verbose=verbose)
            assert self.embedding_cell_types is not None
            assert self.comparison_cell_types is not None
            assert self.original_cell_types is not None
            assert self.comparison_data is not None
            self.evaluations["concordance"] = {"cluster distance": {}, "emd": {}, "gating concordance: ARI": {}, "gating concordance: NMI": {}}
            for e in self.reductions.keys():
                self.evaluations["concordance"]["emd"][e] = EvaluationMetrics.embedding_concordance(self.reductions[e],
                                                                                                    self.embedding_cell_types[e],
                                                                                                    self.comparison_data,
                                                                                                    self.comparison_cell_types,
                                                                                                    self.comparison_classes,
                                                                                                    "emd")
                self.evaluations["concordance"]["cluster distance"][e] = EvaluationMetrics.embedding_concordance(self.reductions[e],
                                                                                                                 self.embedding_cell_types[e],
                                                                                                                 self.comparison_data,
                                                                                                                 self.comparison_cell_types,
                                                                                                                 self.comparison_classes,
                                                                                                                 "cluster_distance")
                self.evaluations["concordance"]["gating concordance: ARI"][e] = EvaluationMetrics.ARI(x_labels=self.embedding_cell_types[e],
                                                                                                      y_labels=self.original_cell_types)
                self.evaluations["concordance"]["gating concordance: NMI"][e] = EvaluationMetrics.NMI(x_labels=self.embedding_cell_types[e],
                                                                                                      y_labels=self.original_cell_types)
          

    def rank_dr_methods(self, tie_method: str="max"):
        """Rank DR Methods Using Default DR Evaluation.

        Based on the results from the ``evaluate`` method, this method ranks the DR methods
        based on the categories chosen. All weighting schemes are consistent with the paper.
        Custom evaluation and weighting schemes are not supported in this case.
        
        :param tie_method: The method to deal with ties when ranking, defaults to "max".
        :type tie_method: str, optional

        :return: A dictionary of DR methods and their final weighted ranks.
        """
        category_counter: int = 0
        overall_rank: np.ndarray = np.zeros(len(self.reductions))
        if "global" in self.evaluations.keys():
            category_counter += 1
            global_eval: np.ndarray = scipy.stats.rankdata(np.array(list(self.evaluations["global"]["spearman"].values())), method=tie_method)/2
            global_eval += scipy.stats.rankdata(-np.array(list(self.evaluations["global"]["emd"].values())), method=tie_method)/2
            overall_rank += global_eval
            
        if "local" in self.evaluations.keys():
            category_counter += 1
            local_eval: np.ndarray = scipy.stats.rankdata(np.array(list(self.evaluations["local"]["knn"].values())), method=tie_method)/2
            local_eval += scipy.stats.rankdata(-np.array(list(self.evaluations["local"]["npe"].values())), method=tie_method)/2
            overall_rank += local_eval
    
        if "downstream" in self.evaluations.keys():
            category_counter += 1
            cluster_reconstruction_eval: np.ndarray = scipy.stats.rankdata(np.array(list(self.evaluations["downstream"]["cluster reconstruction: RF"].values())), method=tie_method)/4
            cluster_reconstruction_eval += scipy.stats.rankdata(np.array(list(self.evaluations["downstream"]["cluster reconstruction: silhouette"].values())), method=tie_method)/4
            cluster_reconstruction_eval += scipy.stats.rankdata(-np.array(list(self.evaluations["downstream"]["cluster reconstruction: DBI"].values())), method=tie_method)/4
            cluster_reconstruction_eval += scipy.stats.rankdata(np.array(list(self.evaluations["downstream"]["cluster reconstruction: CHI"].values())), method=tie_method)/4
            
            cluster_concordance_eval: np.ndarray = scipy.stats.rankdata(np.array(list(self.evaluations["downstream"]["cluster concordance: ARI"].values())), method=tie_method)/2
            cluster_concordance_eval += scipy.stats.rankdata(np.array(list(self.evaluations["downstream"]["cluster concordance: NMI"].values())), method=tie_method)/2
            
            if len(self.evaluations["downstream"]["cell type-clustering concordance: ARI"]) > 0:
                type_cluster_concordance_eval: np.ndarray = scipy.stats.rankdata(np.array(list(self.evaluations["downstream"]["cell type-clustering concordance: ARI"].values())), method=tie_method)/2
                type_cluster_concordance_eval += scipy.stats.rankdata(np.array(list(self.evaluations["downstream"]["cell type-clustering concordance: ARI"].values())), method=tie_method)/2
                downstream_eval: np.ndarray = (cluster_reconstruction_eval + cluster_concordance_eval + type_cluster_concordance_eval)/3
            else:
                downstream_eval: np.ndarray = (cluster_reconstruction_eval + cluster_concordance_eval)/2
            overall_rank += downstream_eval
                
        if "concordance" in self.evaluations.keys():
            category_counter += 1
            concordance_eval: np.ndarray = scipy.stats.rankdata(-np.array(list(self.evaluations["concordance"]["cluster distance"].values())), method=tie_method)/3
            concordance_eval += scipy.stats.rankdata(-np.array(list(self.evaluations["concordance"]["emd"].values())), method=tie_method)/3
            concordance_eval += scipy.stats.rankdata(np.array(list(self.evaluations["concordance"]["gating concordance: ARI"].values())), method=tie_method)/6
            concordance_eval += scipy.stats.rankdata(np.array(list(self.evaluations["concordance"]["gating concordance: NMI"].values())), method=tie_method)/6
            overall_rank += concordance_eval
            
        overall_rank = overall_rank/category_counter
        return dict(zip(self.reductions.keys(), list(overall_rank)))
    
    
    def add_custom_evaluation_result(self, metric_name: str, reduction_name: str, value: float, weight: Optional[float]=None, reverse_ranking: bool=False):
        """Add custom evaluation result.

        This method allows you work with custom evaluation metrics. Instead of relying on the
        builtin schemes, you can add any metrics you would like. This offers great flexibility
        by allowing you to use metrics not included in the ``CytofDR`` package. However, the
        downside is that you have to run these metrics manually in your workflow. This method
        simply stores the results in the object so that the DR methods can automatically ranked.
        
        .. note::

            This method does not distinguish between metrics' weight and revese_ranking properties
            across different embeddings. For example, if you add the same metrics for ``umap`` and
            ``tsne``, they should have the same ``weight`` and ``reverse_ranking`` properties, with the
            only difference being ``value``. Otherwise, latter properties will overwrite previous ones.
        
        :param metric_name: The name of the custom metric.
        :type metric_name: str
        :param reduction_name: The name pf the reduction on which the metric is run.
        :type reduction_name: str
        :param value: The value of the evaluation metric.
        :type value: float
        :param weight: The weight of the metric in the overall DR ranking scheme, defaults to None
        :type weight: Optional[float], optional
        :param reverse_ranking: Whether it should be reverse ranked, defaults to False. If ``True``, this means that smaller values are better.
        :type reverse_ranking: bool, optional
        """
        if metric_name not in self.custom_evaluations["custom"].keys():
            self.custom_evaluations["custom"][metric_name] = {}
        self.custom_evaluations["custom"][metric_name][reduction_name] = value
        
        if weight is not None:
            self.custom_evaluation_weights[metric_name] = weight
        self.custom_evaluation_reverse_ranking[metric_name] = reverse_ranking        


    def rank_dr_methods_custom(self, tie_method: str="max") -> Dict[str, float]:
        """Rank DR methods according to custom evaluation metrics.

        This is the custom version the ``rank_dr_methods`` method. It ranks all the
        DR methods using the metrics added through the ``add_custom_evaluation_result``
        method and their corresonding weights.

        :param tie_method: The method to deal with ties when ranking, defaults to "max"
        :type tie_method: str, optional
        :raises RuntimeError: Missing metrics for certain DR methods.
        :return: A dictionary of methods with their weighted rankings.
        :rtype: Dict[str, float]
        """
        
        custom_evaluations: Dict[str, Any] = {}
        all_dr_methods: np.ndarray = np.array(list(self.names))
        for metric in self.custom_evaluations["custom"].keys():
            method_check: np.ndarray = np.isin(all_dr_methods, np.array(list(self.custom_evaluations["custom"][metric].keys())))
            if not np.all(method_check):
                raise RuntimeError(f"Missing metrics for the following methods: {str(all_dr_methods[np.where(np.invert(method_check))])}")
            custom_evaluations[metric] = dict(sorted(self.custom_evaluations["custom"][metric].items()))
        
        eval_ranks: np.ndarray = np.zeros((all_dr_methods.shape[0],))
        for metric in custom_evaluations.keys():
            rev: int = -1 if self.custom_evaluation_reverse_ranking[metric] else 1
            weight: float = self.custom_evaluation_weights[metric] if len(self.custom_evaluation_weights) != 0 else 1/len(custom_evaluations.keys())
            eval_ranks += scipy.stats.rankdata(rev*np.array(list(custom_evaluations[metric].values())), method=tie_method)*weight
            
        return dict(zip(custom_evaluations[list(custom_evaluations)[0]].keys(), list(eval_ranks)))
                
            
    def plot_reduction(self, name: str, save_path: str, style: str="darkgrid", hue: Optional["np.ndarray"]=None, **kwargs):
        """Draw embedding using a scatter plot.

        This method generates a scatter plot for reductions in the class. 
        
        :param name: The name of the reduction.
        :param save_path: The path to save the plot.
        :param stype: The plot style, defaults to "darkgrid".
        :param hue: Labels used to color the points.
        :param kwargs: Keyword arguments passed into the ``sns.scatterplot`` method.
        
        .. note:: Live viewing is not supported by this method.
        """
        plt.figure()
        sns.set_style(style)
        sns_plot = sns.scatterplot(x=self.reductions[name][:,0], y=self.reductions[name][:,1], hue=hue, **kwargs)
        fig = sns_plot.get_figure()
        fig.savefig(save_path)
        plt.clf()
        
        
    def save_reduction(self, name: str, path: str, overwrite: bool=False, delimiter: str="\t", **kwargs):
        """Save a DR embeddings to disk.

        This method saves a specific reduction to a disk. The file is presumed
        to be a plain text file. All additional arguments are passed to ``np.savetxt``.

        :param name: The name of the embedding to be saved as the key stored in the ``reductions`` dictionary.
        :type name: str
        :param path: The path to save the 
        :type path:
        :param overwrite: Whether to overwrite existing files, defaults to False
        :type overwrite: bool, optional
        :param delimiter: The delimiter used, defaults to "\t"
        :type delimiter: str, optional
        """
        if not overwrite and os.path.exists(path):
            raise FileExistsError()
        np.savetxt(path, self.reductions[name], delimiter=delimiter, **kwargs)
        
        
    def save_all_reductions(self, save_dir: str, overwrite: bool=False, delimiter: str="\t", **kwargs):
        """Save all DR embeddings to a specified directory.

        This method saves all reductions to a specified directory. All files
        are plain text files with their embedding names as their names. All
        additional arguments are passed to ``np.savetxt``.

        :param save_dir: The directory path to save all the embeddings.
        :type save_dir: str
        :param overwrite: Whether to overwrite existing files, defaults to False
        :type overwrite: bool, optional
        :param delimiter: The delimiter used, defaults to "\t"
        :type delimiter: str, optional
        :param kwargs: Keyword-only arguments passed to the `np.savetxt` method.
        """
        for method in self.names:
            path = save_dir + method + ".txt"
            self.save_reduction(method, path, overwrite, delimiter, **kwargs)
            
            
    def save_evaluations(self, path: str, overwrite: bool=False):
        """Save DR evaluation results.

        This method saves all the results of DR evaluation in a
        csv.

        :param path: The path to the file.
        :type path: str
        :param overwrite: Whether to overwrite existing file, defaults to False
        :type overwrite: bool, optional
        :raises AttributeError: There is no ``evaluations`` attribute. Need to run ``evaluate`` first.
        :raises FileExistsError: The file already exists and ``overwrite`` is set to false.
        """
        if len(self.evaluations) == 0:
            raise AttributeError("No 'evaluations' values found. Run the 'evaluate()' method first.")
        if os.path.exists(path) and not overwrite:
            raise FileExistsError()
        
        with open(path, "w") as f:
            w: "_csv._writer" = csv.writer(f)
            w.writerow(["Category", "Metric", "Method", "Value"])
                    
            category: str
            for category in self.evaluations.keys():
                for metric in self.evaluations[category].keys():
                    for method in self.evaluations[category][metric].keys():
                        w.writerow([category, metric, method, self.evaluations[category][metric][method]])
                        
                        
    def __str__(self) -> str:
        """Implementation of `print()`.

        This implementation reductions a description of the object with the names of the embeddings.

        :return: The output string of the `print()` method.
        """
        return f"A 'Reductions' object with {', '.join(self.names)}."
    
    
    def __getitem__(self,  name: str) -> np.ndarray:
        """Get a specific embedding from the ``Reductions`` object.

        This method implements the bracket notation to access reductions. This is
        equivalent to ``get_reduction()``.

        :param name: The name of the reduction.
        :type name: str
        :return: A numpy array of the reduction.
        :rtype: np.ndarray
        """
        return self.reductions[name]
    
    
    def __setitem__(self, name: str, reduction: "np.ndarray"):
        """To set reductions to the ``Reductions`` object.

        This method implements the bracket notation to add reductions
        to the object. This is largely the same as the ``add_reduction()``
        method. The difference is that here there is no overwrite
        protection, but the ``add_reduction()`` method defaults to
        raise an error instead of replacing existing embeddings.

        :param name: The name of the reduction.
        :type name: str
        :param reduction: A numpy array of the reduction.
        :type reduction: np.ndarray
        """
        self.add_reduction(reduction, name, replace=True)
        
        
    def __delitem__(self, name: str):
        """Delete a DR embedding from the object.

        This method implements the `del` keyword and deletes the specified DR embedding.
        The `names` attribute will be reset as well.
        
        .. note:: For now, any downstream rankings and analyses will **not** be reset with this operation.

        :param name: The name of the embedding.
        :type name: str
        """
        del self.reductions[name]
        self.names = set(self.reductions.keys())
        
        
    @property
    def reductions(self) -> Dict[str, "np.ndarray"]:
        """Getter for the `reductions` dictionary.

        This method returns the dictionary storing all the reductions in this class.
        The dictionary keys are names of the embeddings, and its values are the arrays
        of reductions.

        :return: A dictionary of reductions.
        """
        return self._reductions
    
        
    @reductions.setter
    def reductions(self, reductions: Dict[str, "np.ndarray"]):
        self._reductions = reductions
        self.names = set(reductions.keys())
        
    
    
class LinearMethods():
    """Linear DR Methods
    
    This class contains static methods of a group of Linear DR Methods. If
    available, the sklearn implementation is used. All keyword arguments are
    passed directly to the method itself to allow for flexibility.
    """
    
    @staticmethod
    def PCA(data: "np.ndarray",
            out_dims: int=2,
            **kwargs
            ) -> "np.ndarray":
        """Scikit-Learn Principal Component Analysis (PCA)

        This method uses the Sklearn's standard PCA.
        
        :param data: The input high-dimensional array.
        :param out_dims: The number of dimensions of the output, defaults to 2.

        :return: The low-dimensional embedding.
        """
        return PCA(n_components=out_dims, **kwargs).fit_transform(data)
    
    
    @staticmethod
    def ICA(data: "np.ndarray",
            out_dims: int=2,
            **kwargs) -> "np.ndarray":   
        """Scikit-Learn Independent Component Analysis (ICA)

        This method uses the SKlearn's FastICA implementation of ICA.
        
        :param data: The input high-dimensional array.
        :param out_dims: The number of dimensions of the output, defaults to 2.

        :return: The low-dimensional embedding.
        """
        return FastICA(n_components=out_dims, **kwargs).fit_transform(data) #type: ignore
    
    
    @staticmethod
    def ZIFA(data: "np.ndarray",
             out_dims: int=2,
             **kwargs) -> "np.ndarray":
        """Zero-Inflated Factor Analysis (ZIFA)

        This method implements ZIFA as developed by Pierson & Yau (2015).
        
        :param data: The input high-dimensional array.
        :param out_dims: The number of dimensions of the output, defaults to 2.

        :return: The low-dimensional embedding.
        """
        # Fix all-zero columns
        nonzero_col: List[int] = []
        col: int
        for col in range(data.shape[1]):
            if not np.all(data[:,col]==0):
                nonzero_col.append(col)
        data = data[:, nonzero_col]
        
        z: "np.ndarray"
        z, _ = ZIFA.fitModel(data, out_dims, **kwargs)
        
        return z
    
    
    @staticmethod
    def FA(data: "np.ndarray",
           out_dims: int=2,
           **kwargs) -> "np.ndarray":
        """Scikit-Learn Factor Analysis (FA)

        This method uses the SKlearn's FA implementation.
        
        :param data: The input high-dimensional array.
        :param out_dims: The number of dimensions of the output, defaults to 2.

        :return: The low-dimensional embedding.
        """
        return FactorAnalysis(n_components=out_dims, **kwargs).fit_transform(data)
    
    
    @staticmethod
    def NMF(data: "np.ndarray",
            out_dims: int=2,
            **kwargs) -> "np.ndarray":
        """Scikit-Learn Nonnegative Matrix Factorization (NMF)

        This method uses the SKlearn's NMF implementation.
        
        :param data: The input high-dimensional array.
        :param out_dims: The number of dimensions of the output, defaults to 2.

        :return: The low-dimensional embedding.
        """
        return NMF(n_components=out_dims, init = "nndsvd", **kwargs).fit_transform(data)
    
    
class NonLinearMethods():
    """NonLinear DR Methods.

    This class contains static methods of a group of NonLinear DR Methods, except for tSNE.
    """
    
    @staticmethod
    def MDS(data: "np.ndarray",
            out_dims: int=2,
            n_jobs: int=-1,
            **kwargs) -> "np.ndarray":
        """Scikit-Learn Multi-Dimensional Scaling (MDS)

        This method uses the SKlearn's MDS implementation.
        
        :param data: The input high-dimensional array.
        :param out_dims: The number of dimensions of the output, defaults to 2.
        :param n_jobs: The number of jobs to run concurrantly, defaults to -1.

        :return: The low-dimensional embedding.
        """
        return MDS(n_components=out_dims,
                   n_jobs=n_jobs,
                   **kwargs
                   ).fit_transform(data)
    
    
    @staticmethod
    def UMAP(data: "np.ndarray",
             out_dims: int=2,
             n_jobs: int=-1,
             **kwargs
             ) -> "np.ndarray":
        """UMAP

        This method uses the UMAP package's UMAP implementation.
        
        :param data: The input high-dimensional array.
        :param out_dims: The number of dimensions of the output, defaults to 2.
        :param n_jobs: The number of jobs to run concurrantly, defaults to -1.

        :return: The low-dimensional embedding.
        """
        return umap.UMAP(n_components=out_dims,
                         n_jobs=n_jobs,
                         **kwargs
                         ).fit_transform(data) #type: ignore
    
    
    @staticmethod
    def SAUCIE(data: "np.ndarray",
               steps: int=1000,
               batch_size: int=256,
               **kwargs
               ) -> "np.ndarray":
        """SAUCIE

        This method is a wrapper for SAUCIE package's SAUCIE model. Specifically,
        dimension reduction is of interest. Here, all keyword arguments are passed
        into the ``SAUCIE.SAUCIE`` method. The training parameters ``steps`` and 
        ``batch_size`` are directly exposed in this wrapper.
        
        :param data: The input high-dimensional array.
        :param steps: The number of training steps to use, defaults to 1000.
        :param batch_size: The batch size for training, defaults to 256.

        :return: The low-dimensional embedding.
        """
        
        saucie: "SAUCIE.model.SAUCIE" = SAUCIE.SAUCIE(data.shape[1], **kwargs)
        train: "SAUCIE.loader.Loader" = SAUCIE.Loader(data, shuffle=True)
        saucie.train(train, steps=steps, batch_size=batch_size)
        
        eval: "SAUCIE.loader.Loader" = SAUCIE.Loader(data, shuffle=False)
        embedding: "np.ndarray" = saucie.get_embedding(eval) #type: ignore
        
        return embedding
    
    
    @staticmethod
    def isomap(data: "np.ndarray",
               out_dims: int=2,
               transform: Optional["np.ndarray"]=None,
               n_jobs: int=-1,
               **kwargs
               ) -> "np.ndarray":
        """Scikit-Learn Isomap

        This method is a wrapper for sklearn's implementation of Isomap.
        
        :param data: The input high-dimensional array.
        :param out_dims: The number of dimensions of the output, defaults to 2.
        :param transform: The array to transform with the trained model.
        :param n_jobs: The number of threads to use.

        :return: The low-dimensional embedding.
        """
        
        if transform is None:
            embedding: "np.ndarray" = Isomap(n_components=out_dims,
                                             n_jobs=n_jobs,
                                             **kwargs).fit_transform(data)
        else:
            embedding: "np.ndarray" = Isomap(n_components=out_dims,
                                             n_jobs=n_jobs,
                                             **kwargs).fit(data).transform(transform)
        
        return embedding
    
    
    @staticmethod
    def LLE(data: "np.ndarray",
            out_dims: int=2,
            transform: Optional["np.ndarray"]=None,
            n_jobs: int=-1,
            **kwargs
            ) -> "np.ndarray": 
        """Scikit-Learn Locally Linear Embedding (LLE)

        This method is a wrapper for sklearn's implementation of LLE.
        
        :param data: The input high-dimensional array.
        :param out_dims: The number of dimensions of the output, defaults to 2.
        :param transform: The array to transform with the trained model.

        :return: The low-dimensional embedding.
        """
        
        if transform is None:
            embedding: "np.ndarray" = LocallyLinearEmbedding(n_components=out_dims,
                                                             n_jobs=n_jobs,
                                                             **kwargs).fit_transform(data)
        else:
            embedding: "np.ndarray" = LocallyLinearEmbedding(n_components=out_dims,
                                                             n_jobs=n_jobs,
                                                             **kwargs).fit(data).transform(transform)
        
        return embedding
    
    
    @staticmethod
    def kernelPCA(data: "np.ndarray",
                  out_dims: int=2,
                  kernel: str="poly",
                  n_jobs: int=-1,
                  **kwargs
                  ) -> "np.ndarray": 
        """Scikit-Learn Kernel PCA

        This method is a wrapper for sklearn's implementation of kernel PCA.
        
        :param data: The input high-dimensional array.
        :param out_dims: The number of dimensions of the output, defaults to 2.
        :param kernel: The kernel to use: "poly," "linear," "rbf," "sigmoid," or "cosine."
        :param n_jobs: The number of jobs to run concurrantly, defaults to -1.

        :return: The low-dimensional embedding.
        """
        return KernelPCA(n_components=out_dims,
                         kernel=kernel,
                         n_jobs=n_jobs,
                         **kwargs).fit_transform(data)


    @staticmethod
    def spectral(data: "np.ndarray",
                 out_dims: int=2,
                 n_jobs: int=-1,
                 **kwargs):
        """Scikit-Learn Spectral Embedding

        This method is a wrapper for sklearn's implementation of spectral embedding.
        
        :param data: The input high-dimensional array.
        :param out_dims: The number of dimensions of the output, defaults to 2.
        :param n_jobs: The number of jobs to run concurrantly, defaults to -1.

        :return: The low-dimensional embedding.
        """
        return SpectralEmbedding(n_components=out_dims,
                                 n_jobs=n_jobs,
                                 **kwargs).fit_transform(data)

    
    @staticmethod
    def phate(data: "np.ndarray",
              out_dims: int=2,
              n_jobs: int=-1,
              **kwargs):
        """PHATE

        This method is a wrapper for PHATE.
        
        :param data: The input high-dimensional array.
        :param out_dims: The number of dimensions of the output, defaults to 2.
        :param n_jobs: The number of jobs to run concurrantly, defaults to -1.

        :return: The low-dimensional embedding.
        """
        return phate.PHATE(n_components=out_dims,
                           n_jobs=n_jobs,
                           **kwargs).fit_transform(data)
    
    
    @staticmethod
    def grandprix(data: "np.ndarray",
                  out_dims: int=2,
                  **kwargs):
        """GrandPrix

        This method is a wrapper for GrandPrix.
        
        :param data: The input high-dimensional array.
        :param out_dims: The number of dimensions of the output, defaults to 2.

        :return: The low-dimensional embedding.
        """
        return GrandPrix.fit_model(data = data, n_latent_dims = out_dims, **kwargs)[0]
    
    
    @staticmethod
    def sklearn_tsne(data: "np.ndarray",
                     out_dims: int=2,
                     n_jobs: int=-1,
                     **kwargs
                     )-> "np.ndarray":
        """Scikit-Learn t-SNE

        This method uses the Scikit-learn implementation of t-SNE. It supports both
        traditional and BH t-SNE with more control of variables.
        
        :param data: The input high-dimensional array.
        :param out_dims: The number of dimensions of the output, defaults to 2.
        :param n_jobs: The number of jobs to run concurrantly, defaults to -1.

        :return: The low-dimensional embedding.
        """
        return TSNE(n_components=out_dims,
                    n_jobs=n_jobs,
                    **kwargs
                    ).fit_transform(data)
    
    
    @staticmethod
    def open_tsne(data: "np.ndarray",
                  out_dims: int=2,
                  perp: Union[List[int],int]=30,
                  learning_rate: Union[str, float]="auto",
                  early_exaggeration_iter: int=250,
                  early_exaggeration: float=12,
                  max_iter: int=500,
                  metric: str="euclidean",
                  dof: int=1,
                  theta: float=0.5, 
                  init: Union["np.ndarray", str]="pca",
                  negative_gradient_method: str="fft",
                  n_jobs: int=-1
                  ) -> "np.ndarray":
        """openTSNE implementation of FIt-SNE

        This is the Python implementation of FIt-SNE through the ``openTSNE`` package. Its implementation
        is based on research from Linderman et al. (2019). This is the default recommended implementation.
        To allow for flexibility and avoid confusion, common parameters are directly exposed without allowing
        additional keyword arguments. 
        
        :param data: The input high-dimensional array.
        :param out_dims: The number of dimensions of the output, defaults to 2.
        :param perp: Perplexity. The default is set to 30. Tradition is between 30 and 50.
            This also supports multiple perplexities with a list, defaults to 30.
        :param learning_rate: The learning rate used during gradient descent, defaults to "auto".
        :param early_exaggeration_iter: Number of early exaggeration iterations, defaults to 250.
        :param early_exaggeration: Early exaggeration factor, defaults to 12.
        :param max_iter: Maximum number of iterations to optimize, defaults to 500
        :param dof: T-distribution degree of freedom, defaults to "euclidean"
        :param theta: The speed/accuracy trade-off, defaults to 0.5.
        :param init: Method of initialiazation. 'random', 'pca', 'spectral', or array, defaults to "pca"
        :param negative_gradient_method: Whether to use "bh" or "fft" tSNE, defaults to "fft".
        :param n_jobs: The number of jobs to run concurrantly, defaults to -1.
        
        :return: The low-dimensional embedding.
        """
        
        n_iter: int = max_iter - early_exaggeration_iter
        affinities_array: Union["affinity.PerplexityBasedNN", "affinity.Multiscale"]
        init_array: "np.ndarray" = np.empty((data.shape[0], out_dims))
        
        if isinstance(perp, list) and len(perp) > 1:
            affinities_array = affinity.Multiscale(data=data,
                                                   perplexities=perp,
                                                   metric=metric,
                                                   n_jobs=n_jobs,
                                                   verbose=True)
        else:
            perp = perp[0] if isinstance(perp, list) else perp
            affinities_array = affinity.PerplexityBasedNN(data=data,
                                                          perplexity=perp,
                                                          metric=metric,
                                                          n_jobs=n_jobs,
                                                          verbose=True)

        if isinstance(init, str):
            if init == "pca":
                init_array = initialization.pca(data, n_components=out_dims)
            elif init == "spectral":
                init_array = initialization.spectral(affinities_array.P, n_components=out_dims)
            else:
                init_array = initialization.random(data, n_components=out_dims)
        else:
            init_array = init
        
        embedding: "np.ndarray" = TSNEEmbedding(embedding=init_array,
                                                affinities=affinities_array,
                                                negative_gradient_method=negative_gradient_method,
                                                learning_rate=learning_rate,
                                                theta=theta,
                                                dof=dof,
                                                n_jobs=n_jobs,
                                                verbose=True)
        # Early exaggeration
        embedding.optimize(n_iter=early_exaggeration_iter,
                           inplace=True,
                           exaggeration=early_exaggeration,
                           momentum=0.5)
        # Regular GD
        embedding.optimize(n_iter=n_iter,
                           inplace=True,
                           momentum=0.8)
        
        return embedding    
    

def run_dr_methods(data: "np.ndarray",
                   methods: Union[str, List[str]]="all",
                   out_dims: int=2,
                   transform: Optional["np.ndarray"]=None,
                   n_jobs: int=-1,
                   verbose: bool=True,
                   suppress_error_msg: bool=False
                   ) -> "Reductions":
    """Run dimension reduction methods.

    This is a one-size-fits-all dispatcher that runs all supported methods in the module. It
    supports running multiple methods at the same time at the sacrifice of some more
    granular control of parameters. If you would like more customization, run each method
    indicidually instead.
    
    :param data: The input high-dimensional array.
    :param methods: DR methods to run (not case sensitive).
    :param out_dims: Output dimension of DR.
    :param transform: An array to transform after training on the traning set.
    :param n_jobs: The number of jobs to run when applicable, defaults to -1.
    :param verbose: Whether to print out progress, defaults to ``True``.
    :param suppress_error_msg: Whether to suppress error messages print outs, defaults to ``False``.

    :return: A Reductions object with all dimension reductions.
    """
    
    if not isinstance(methods, list):
        methods = [methods]
    methods = [each_method.lower() for each_method in methods]
    
    if "all" in methods:
        methods = ["pca", "ica", "umap", "sklearn_tsne", "open_tsne", "fa", "isomap", "mds", "lle",
                   "kpca_poly", "kpca_rbf", "phate", "nmf", "spectral"]
        if METHODS["SAUCIE"]:
            methods.append("saucie")
        if METHODS["ZIFA"]:
            methods.append("zifa")
        if METHODS["GrandPrix"]:
            methods.append("grandprix")
            
    reductions: Reductions = Reductions()
    reductions.add_evaluation_metadata(original_data=data)
    
    if "pca" in methods:
        try:
            _verbose("Running PCA", verbose=verbose)
            reductions.add_reduction(LinearMethods.PCA(data, out_dims=out_dims), "PCA")
        except Exception as method_error:
            _verbose(str(method_error), verbose=not suppress_error_msg)
            
    if "ica" in methods:
        try:
            _verbose("Running ICA", verbose=verbose)
            reductions.add_reduction(LinearMethods.ICA(data, out_dims=out_dims), "ICA") 
        except Exception as method_error:
            _verbose(str(method_error), verbose=not suppress_error_msg)
        
    if "umap" in methods:
        try:
            _verbose("Running UMAP", verbose=verbose)    
            reductions.add_reduction(NonLinearMethods.UMAP(data, out_dims=out_dims, n_jobs=n_jobs), "UMAP")
        except Exception as method_error:
            _verbose(str(method_error), verbose=not suppress_error_msg)
            
    if "saucie" in methods:
        try:
            _verbose("Running SAUCIE", verbose=verbose)
            reductions.add_reduction(NonLinearMethods.SAUCIE(data), "SAUCIE")
        except Exception as method_error:
            _verbose(str(method_error), verbose=not suppress_error_msg)
    
    # sklearn BH
    if "sklearn_tsne" in methods:
        try:
            _verbose("Running sklearn_tsne", verbose=verbose)
            reductions.add_reduction(NonLinearMethods.sklearn_tsne(data, out_dims=out_dims, n_jobs=n_jobs), "sklearn_tsne")
        except Exception as method_error:
            _verbose(str(method_error), verbose=not suppress_error_msg)
    
    if "open_tsne" in methods:
        try:
            _verbose("Running open_tsne", verbose=verbose)
            reductions.add_reduction(NonLinearMethods.open_tsne(data, out_dims=out_dims, n_jobs=n_jobs), "open_tsne")
        except Exception as method_error:
            _verbose(str(method_error), verbose=not suppress_error_msg)
            
    if "zifa" in methods:
        try:
            _verbose("Running ZIFA", verbose=verbose)
            reductions.add_reduction(LinearMethods.ZIFA(data, out_dims=out_dims), "ZIFA")
        except Exception as method_error:
            _verbose(str(method_error), verbose=not suppress_error_msg)
            
    if "fa" in methods:
        try:
            _verbose("Running FA", verbose=verbose)
            reductions.add_reduction(LinearMethods.FA(data, out_dims=out_dims), "FA")
        except Exception as method_error:
            _verbose(str(method_error), verbose=not suppress_error_msg)
            
    if "isomap" in methods:
        try:
            _verbose("Running Isomap", verbose=verbose)
            reductions.add_reduction(NonLinearMethods.isomap(data, out_dims=out_dims,transform=transform, n_jobs=n_jobs), "Isomap")
        except Exception as method_error:
            _verbose(str(method_error), verbose=not suppress_error_msg)
            
    if "mds" in methods:
        try:
            _verbose("Running MDS", verbose=verbose)
            reductions.add_reduction(NonLinearMethods.MDS(data, out_dims=out_dims, n_jobs=n_jobs), "MDS")
        except Exception as method_error:
            _verbose(str(method_error), verbose=not suppress_error_msg)
            
    if "lle" in methods:
        try:
            _verbose("Running LLE", verbose=verbose)
            reductions.add_reduction(NonLinearMethods.LLE(data, out_dims=out_dims, transform=transform, n_jobs=n_jobs), "LLE")
        except Exception as method_error:
            _verbose(str(method_error), verbose=not suppress_error_msg)
            
    if "spectral" in methods:
        try:
            _verbose("Running Spectral", verbose=verbose)
            reductions.add_reduction(NonLinearMethods.spectral(data, out_dims=out_dims, n_jobs=n_jobs), "Spectral")
        except Exception as method_error:
            _verbose(str(method_error), verbose=not suppress_error_msg)
                    
    if "kpca_poly" in methods:
        try:
            _verbose("Running kpca_poly", verbose=verbose)
            reductions.add_reduction(NonLinearMethods.kernelPCA(data, out_dims=out_dims, kernel="poly", n_jobs=n_jobs), "kpca_poly")
        except Exception as method_error:
            _verbose(str(method_error), verbose=not suppress_error_msg)
            
    if "kpca_rbf" in methods:
        try:
            _verbose("Running kpca_rbf", verbose=verbose)
            reductions.add_reduction(NonLinearMethods.kernelPCA(data, out_dims=out_dims, kernel="rbf", n_jobs=n_jobs), "kpca_rbf")
        except Exception as method_error:
            _verbose(str(method_error), verbose=not suppress_error_msg)
            
    if "phate" in methods:
        try:
            _verbose("Running PHATE", verbose=verbose)
            reductions.add_reduction(NonLinearMethods.phate(data, out_dims=out_dims, n_jobs=n_jobs), "PHATE")
        except Exception as method_error:
            _verbose(str(method_error), verbose=not suppress_error_msg)
            
    if "nmf" in methods:
        try:
            _verbose("Running nmf", verbose=verbose)
            reductions.add_reduction(LinearMethods.NMF(data, out_dims=out_dims), "NMF")
        except Exception as method_error:
            _verbose(str(method_error), verbose=not suppress_error_msg)
            
    if "grandprix" in methods:
        try:
            _verbose("Running grandprix", verbose=verbose)
            reductions.add_reduction(NonLinearMethods.grandprix(data, out_dims=out_dims), "GrandPrix")
        except Exception as method_error:
            _verbose(str(method_error), verbose=not suppress_error_msg)
            
    return reductions
